#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <malloc.h>
#include <string.h>
#include "lab.h"



static char
toHexa(
	uint8	nombre
)
{
	static const char	table[] = {
		'0', '1', '2', '3', '4', '5', '6', '7',
		'8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
	};
	assert(nombre<=15);
	return	table[nombre];
}


static uint8
toByte(
	char	hexa
)
{
	if	(hexa>='0'&&hexa<='9')
		return (uint8)(hexa-'0');
	else if	(hexa>='A'&&hexa<='F')
		return (uint8)(hexa+10-'A');
	else if	(hexa>='a'&&hexa<='f')
		return (uint8)(hexa+10-'a');
	return	0;
}


lab*
lab_alloueMemoire(
	uint8	largeur,
	uint8	hauteur
)
{
	lab	*tmp;
	uint16	taille;

	assert(largeur>=3&&largeur<=79&&hauteur>=3&&hauteur<=23);
	if	(!(tmp = (lab*)malloc(sizeof(lab))))
		return	(lab*)NULL;

	taille = ((uint16)largeur)*((uint16)hauteur)*sizeof(char);
	tmp->h = hauteur;
	tmp->l = largeur;
	if	(!(tmp->donnees=(uint8*)malloc(taille)))
	{
		free((void*)tmp);
		return	(lab*)NULL;
	}
	memset(tmp->donnees, 0, taille);

	return	tmp;
}


void
lab_detruit(
	lab	*l
)
{
	assert(l&&l->donnees);
	free(l->donnees);
	free(l);
}


uint8
lab_sauvegarde(
	const lab	*l,
	const char	*nomFichier
)
{
	FILE	*f;
	char	*buffer;
	uint16	tmp;

	assert(nomFichier&&l);

	if	(!(f = fopen(nomFichier, "w")))
		return	0;

	tmp = (5+(((uint16)l->h)*((uint16)l->l)))*sizeof(char);
	if	(!(buffer=malloc(tmp)))
	{
		fclose(f);
		return	0;
	}

	buffer[0] = toHexa(l->l/16);
	buffer[1] = toHexa(l->l%16);
	buffer[2] = toHexa(l->h/16);
	buffer[3] = toHexa(l->h%16);

	for	(tmp=0;tmp<(((uint16)l->l)*((uint16)l->h));tmp++)
		buffer[tmp+4] = toHexa(l->donnees[tmp]);
	buffer[tmp+4] = '0';

	fprintf(f, "%s\n", buffer);

	fclose(f);
	free(buffer);

	return	1;
}


lab*
lab_charge(
	const char	*nomFichier
)
{
	FILE	*f;
	char	buffer[2];
	lab	*tmp;
	uint8	largeur, hauteur;
	uint16	i, tot;

	// Ouverture du fichier
	assert(nomFichier);
	if	(!(f=fopen(nomFichier, "r")))
		return	(lab*)NULL;

	// Lecture de la largeur
	if	((buffer[0] = fgetc(f))==EOF)
	{
		fclose(f);
		return	(lab*)NULL;
	}
	if	((buffer[1] = fgetc(f))==EOF)
	{
		fclose(f);
		return	(lab*)NULL;
	}
	largeur = (toByte(buffer[0]) << 4) | toByte(buffer[1]);

	// Lecture de la hauteur
	if	((buffer[0] = fgetc(f))==EOF)
	{
		fclose(f);
		return	(lab*)NULL;
	}
	if	((buffer[1] = fgetc(f))==EOF)
	{
		fclose(f);
		return	(lab*)NULL;
	}
	hauteur = (toByte(buffer[0]) << 4) | toByte(buffer[1]);

	// Allocation de la mmoire
	if	(!(largeur>=3&&largeur<=79&&hauteur>=3&&hauteur<=23))
	{
		fclose(f);
		return	(lab*)NULL;
	}
	tmp = lab_alloueMemoire(largeur, hauteur);

	// Lecture des donnes
	tot = ((uint16)largeur)*((uint16)hauteur);
	for	(i=0;i<tot;i++)
	{
		uint8	valeur;

		buffer[0] = fgetc(f);
		if	(buffer[0]==EOF)
		{
			lab_detruit(tmp);
			fclose(f);
			return	(lab*)NULL;
		}

		valeur = toByte(buffer[0]);
		valeur &= 0x03;
		tmp->donnees[i] = valeur;
	}

	fclose(f);
	return	tmp;
}


void
lab_modifieCase(
	lab	*l,
	uint8	x,
	uint8	y,
	uint8	flagsMurs
)
{
	uint16	position;

	assert(l);
	assert(x<l->l&&y<l->h);
	assert(flagsMurs<=3);

	position = ((uint16)y)*((uint16)l->l) + ((uint16)x);
	l->donnees[position] = flagsMurs;
}


uint8
lab_murEnHaut(
	lab	*l,
	uint8	x,
	uint8	y
)
{
	uint16	position;

	assert(l);
	assert(x<l->l&&y<l->h);

	position = ((uint16)y)*((uint16)l->l) + ((uint16)x);
	return	((l->donnees[position]&MUR_HAUT)!=0);
}


uint8
lab_murAGauche(
	lab	*l,
	uint8	x,
	uint8	y
)
{
	uint16	position;

	assert(l);
	assert(x<l->l&&y<l->h);

	position = ((uint16)y)*((uint16)l->l) + ((uint16)x);
	return	((l->donnees[position]&MUR_GAUCHE)!=0);
}


uint8
lab_affiche(
	lab	*l
)
{
	int	i, j;

	assert(l);
	
	for	(i=0;i<l->h;i++)
	{
		// Affichage des murs du haut pour la ligne concerne
		for	(j=0;j<l->l;j++)
		{
			if	(i==0)
			{
				if	(j==0)
					printf("E ");
				else
					printf("XX");
			}
			else
			{
				if	(lab_murEnHaut(l, j, i))
					printf("XX");
				else
					printf("X ");
			}
		}
		printf("X\n");

		// Affichage des murs gauches
		for	(j=0;j<l->l;j++)
		{
			if	(j==0)
			{
				if	(i==0)
					printf("  ");
				else if	(	lab_murEnHaut(l, j, i)
					&&	(	i==l->h-1
						||	lab_murEnHaut(
								l, j, i+1
							)
						)
					&&	(	j==l->l-1
						||	lab_murAGauche(
								l, j+1, i
							)
						)
					)
					printf("XX");
				else
					printf("X ");
			}
			else if	(lab_murAGauche(l, j, i))
			{
				printf("X");

				// On regarde si cette case est ferme
				if	(	lab_murEnHaut(l, j, i)
					&&	(	i==l->h-1
						||	lab_murEnHaut(
								l, j, i+1
							)
						)
					&&	(	j==l->l-1
						||	lab_murAGauche(
								l, j+1, i
							)
						)
					)
					printf("X");
				else
					printf(" ");
			}
			else
				printf("  ");
		}
		if	(i==l->h-1)
			printf(" \n");
		else
			printf("X\n");
	}

	// Affichage de la ligne du bas
	for	(i=0;i<l->l;i++)
	{
		if	(i<l->l-1)
			printf("XX");
		else
			printf("X S\n");
	}
}


#define	rn(x)	(rand()%(x))

#define DIR_NORD	0
#define	DIR_EST		1
#define	DIR_SUD		2
#define	DIR_OUEST	3


#define dirPrec	((positionHist>0)?dHisto[positionHist-1]:DIR_SUD)

static uint8
fabriqueLaby(
	lab	*l,
	uint8	*genere,
	uint16	xIni,
	uint16	yIni,
	uint16	taille
)
{
	uint16	x, y;
	uint8	essais[4] = { 'N', 'N', 'N', 'N' };
	uint16	*xHisto, *yHisto;
	uint8	*dHisto;
	uint16	positionHist;
	uint8	direction;

	x = xIni;
	y = yIni;
	genere[y*((uint16)l->l)+x] = 1;
	if	(!(xHisto=malloc(taille*2)))
		return;
	if	(!(yHisto=malloc(taille*2)))
		return;
	if	(!(dHisto=malloc(taille)))
		return;
	positionHist = 0;

	while	(!(x==(l->l-1)&&y==(l->h-1)))
	{
		if	(!strncmp(essais, "OOOO", 4))
		{
			// On a dj essay les 4 directions, donc on est
			// bloqu; on revient en arrire
			if	(!(positionHist>0))
				return	0;
			positionHist--;
			x = xHisto[positionHist];
			y = yHisto[positionHist];
			memset(essais, 'N', 4);
		}

		direction = rn(4);
		if	(essais[direction]=='O')
			continue;
		essais[direction] = 'O';
		switch(direction)
		{
			case	DIR_NORD:
				// On vrifie
				if	(dirPrec==DIR_SUD||y==0)
					continue;
				if	(genere[(y-1)*(uint16)l->l+x])
					continue;

				// On enlve le mur
				l->donnees[y*(uint16)l->l+x] &= ~MUR_HAUT;

				// Mise  jour de l'historique
				xHisto[positionHist] = x;
				yHisto[positionHist] = y;
				dHisto[positionHist] = direction;
				positionHist++;

				y--;
				break;
				
			case	DIR_EST:
				// On vrifie
				if	(dirPrec==DIR_OUEST||x==l->l-1)
					continue;
				if	(genere[y*(uint16)l->l+x+1])
					continue;

				// On enlve le mur
				l->donnees[y*(uint16)l->l+x+1] &= ~MUR_GAUCHE;

				// Mise  jour de l'historique
				xHisto[positionHist] = x;
				yHisto[positionHist] = y;
				dHisto[positionHist] = direction;
				positionHist++;

				x++;
				break;
				
			case	DIR_SUD:
				// On vrifie
				if	(dirPrec==DIR_NORD||y==l->h-1)
					continue;
				if	(genere[(y+1)*(uint16)l->l+x])
					continue;

				// On enlve le mur
				l->donnees[(y+1)*(uint16)l->l+x] &= ~MUR_HAUT;

				// Mise  jour de l'historique
				xHisto[positionHist] = x;
				yHisto[positionHist] = y;
				dHisto[positionHist] = direction;
				positionHist++;

				y++;
				break;
				
			case	DIR_OUEST:
				// On vrifie
				if	(dirPrec==DIR_EST||x==0)
					continue;
				if	(genere[y*(uint16)l->l+x-1])
					continue;

				// On enlve le mur
				l->donnees[y*(uint16)l->l+x] &= ~MUR_GAUCHE;

				// Mise  jour de l'historique
				xHisto[positionHist] = x;
				yHisto[positionHist] = y;
				dHisto[positionHist] = direction;
				positionHist++;

				x--;
				break;
		}
		genere[y*(uint16)l->l+x] = 1;
		memset(essais, 'N', 4);
	}
	return	1;
}

#undef dirPrec


void
lab_generer(
	lab	*l
)
{
	uint16	x, y, i, j;
	uint16	taille;
	uint8	*genere;
	uint16	offset;

	assert(l);
	srand(time(0));

	taille = ((uint16)l->l)*((uint16)l->h);
	if	(!(genere=malloc(taille)))
		return;
			
	x = y = 0;
	memset(l->donnees, MUR_HAUT|MUR_GAUCHE, taille);
	memset(genere, 0, taille);

	if	(!fabriqueLaby(l, genere, 0, 0, taille))
	{
		lab_affiche(l);
		exit(-1);
	}

	// Nous avons trac un chemin jusqu' la sortie, mais certaines zones
	// n'auront pas t gnres du tout; on les recherche et on les gnre
	for	(offset=i=0;i<l->h;i++)
	{
		for	(j=0;j<l->l;j++,offset++)
		{
			if	(!genere[offset])
			{
				// Zone non gnre trouve
				l->donnees[offset] &= ~MUR_GAUCHE;
				fabriqueLaby(l, genere, j, i, taille);
			}
		}
	}

	free(genere);
}
